import numpy as np
from rbc.irtools import printf
from rbc import typesystem
from .omnisci_array import Array, ArrayPointer
from rbc.utils import get_version
from numba import njit
if get_version('numba') >= (0, 49):
    from numba.core import extending, types, \
        errors
    from numba.np import numpy_support
else:
    from numba import extending, types, \
        errors, numpy_support


def expose_and_overload(func):
    name = func.__name__
    s = f'def {name}(*args, **kwargs): pass'
    exec(s, globals())

    fn = globals()[name]
    decorate = extending.overload(fn)

    def wrapper(overload_func):
        return decorate(overload_func)

    return wrapper


@expose_and_overload(np.full)
def omnisci_np_full(shape, fill_value, dtype=None):

    # XXX: dtype should be infered from fill_value
    if dtype is None:
        nb_dtype = types.double
    else:
        nb_dtype = typesystem.Type.fromobject(dtype).tonumba()

    def impl(shape, fill_value, dtype=None):
        a = Array(shape, nb_dtype)
        a.fill(nb_dtype(fill_value))
        return a
    return impl


@expose_and_overload(np.full_like)
def omnisci_np_full_like(a, fill_value, dtype=None):
    if isinstance(a, ArrayPointer):
        if dtype is None:
            nb_dtype = a.eltype
        else:
            nb_dtype = typesystem.Type.fromobject(dtype).tonumba()

        def impl(a, fill_value, dtype=None):
            sz = len(a)
            other = Array(sz, nb_dtype)
            other.fill(nb_dtype(fill_value))
            return other
        return impl


@expose_and_overload(np.empty_like)
def omnisci_np_empty_like(a, dtype=None):
    if isinstance(a, ArrayPointer):
        if dtype is None:
            nb_dtype = a.eltype
        else:
            nb_dtype = typesystem.Type.fromobject(dtype).tonumba()

        def impl(a, dtype=None):
            other = Array(0, nb_dtype)
            other.set_null()
            return other
        return impl


@expose_and_overload(np.empty)
def omnisci_np_empty(shape, dtype=None):

    if dtype is None:
        nb_dtype = types.double
    else:
        nb_dtype = typesystem.Type.fromobject(dtype).tonumba()

    def impl(shape, dtype=None):
        arr = Array(shape, nb_dtype)
        arr.set_null()
        return arr
    return impl


@expose_and_overload(np.zeros)
def omnisci_np_zeros(shape, dtype=None):

    if dtype is None:
        nb_dtype = types.double
    else:
        nb_dtype = typesystem.Type.fromobject(dtype).tonumba()

    fill_value = False if isinstance(nb_dtype, types.Boolean) else 0

    def impl(shape, dtype=None):
        return full(shape, fill_value, nb_dtype)  # noqa: F821
    return impl


@expose_and_overload(np.zeros_like)
def omnisci_np_zeros_like(a, dtype=None):
    if isinstance(a, ArrayPointer):
        if dtype is None:
            nb_dtype = a.eltype
        else:
            nb_dtype = typesystem.Type.fromobject(dtype).tonumba()

        fill_value = False if isinstance(nb_dtype, types.Boolean) else 0

        def impl(a, dtype=None):
            return full_like(a, fill_value, nb_dtype)  # noqa: F821
        return impl


@expose_and_overload(np.ones)
def omnisci_np_ones(shape, dtype=None):

    if dtype is None:
        nb_dtype = types.double
    else:
        nb_dtype = typesystem.Type.fromobject(dtype).tonumba()

    fill_value = True if isinstance(nb_dtype, types.Boolean) else 1

    def impl(shape, dtype=None):
        return full(shape, fill_value, nb_dtype)  # noqa: F821
    return impl


@expose_and_overload(np.ones_like)
def omnisci_np_ones_like(a, dtype=None):
    if isinstance(a, ArrayPointer):
        if dtype is None:
            nb_dtype = a.eltype
        else:
            nb_dtype = dtype

        fill_value = True if isinstance(nb_dtype, types.Boolean) else 1

        def impl(a, dtype=None):
            return full_like(a, fill_value, nb_dtype)  # noqa: F821
        return impl


@expose_and_overload(np.array)
def omnisci_np_array(a, dtype=None):

    @njit
    def omnisci_array_non_empty_copy(a, nb_dtype):
        """Implement this here rather than inside "impl".
        LLVM DCE pass removes everything if we implement stuff inside "impl"
        """
        other = Array(len(a), nb_dtype)
        for i in range(len(a)):
            other[i] = a[i]
        return other

    if isinstance(a, ArrayPointer):
        if dtype is None:
            nb_dtype = a.eltype
        else:
            nb_dtype = dtype

        def impl(a, dtype=None):
            if a.is_null():
                return empty_like(a)  # noqa: F821
            else:
                return omnisci_array_non_empty_copy(a, nb_dtype)
        return impl


def get_type_limits(eltype):
    np_dtype = numpy_support.as_dtype(eltype)
    if isinstance(eltype, types.Integer):
        return np.iinfo(np_dtype)
    elif isinstance(eltype, types.Float):
        return np.finfo(np_dtype)
    else:
        msg = 'Type {} not supported'.format(eltype)
        raise errors.TypingError(msg)


@extending.overload_method(ArrayPointer, 'fill')
def omnisci_array_fill(x, v):
    if isinstance(x, ArrayPointer):
        def impl(x, v):
            for i in range(len(x)):
                x[i] = v
        return impl


@extending.overload(max)
@expose_and_overload(np.max)
@extending.overload_method(ArrayPointer, 'max')
def omnisci_array_max(x, initial=None):
    if isinstance(x, ArrayPointer):
        min_value = get_type_limits(x.eltype).min

        def impl(x, initial=None):
            if len(x) <= 0:
                printf("omnisci_array_max: cannot find max of zero-sized array")  # noqa: E501
                return min_value
            if initial is not None:
                m = initial
            else:
                m = x[0]
            for i in range(len(x)):
                v = x[i]
                if v > m:
                    m = v
            return m
        return impl


@extending.overload(min)
@extending.overload_method(ArrayPointer, 'min')
def omnisci_array_min(x, initial=None):
    if isinstance(x, ArrayPointer):
        max_value = get_type_limits(x.eltype).max

        def impl(x, initial=None):
            if len(x) <= 0:
                printf("omnisci_array_min: cannot find min of zero-sized array")  # noqa: E501
                return max_value
            if initial is not None:
                m = initial
            else:
                m = x[0]
            for i in range(len(x)):
                v = x[i]
                if v < m:
                    m = v
            return m
        return impl


@extending.overload(sum)
@expose_and_overload(np.sum)
@extending.overload_method(ArrayPointer, 'sum')
def omnisci_np_sum(a, initial=None):
    if isinstance(a, ArrayPointer):
        def impl(a, initial=None):
            if initial is not None:
                s = initial
            else:
                s = 0
            n = len(a)
            for i in range(n):
                s += a[i]
            return s
        return impl


@expose_and_overload(np.prod)
@extending.overload_method(ArrayPointer, 'prod')
def omnisci_np_prod(a, initial=None):
    if isinstance(a, ArrayPointer):
        def impl(a, initial=None):
            if initial is not None:
                s = initial
            else:
                s = 1
            n = len(a)
            for i in range(n):
                s *= a[i]
            return s
        return impl


@expose_and_overload(np.mean)
@extending.overload_method(ArrayPointer, 'mean')
def omnisci_array_mean(x):
    if isinstance(x.eltype, types.Integer):
        zero_value = 0
    elif isinstance(x.eltype, types.Float):
        zero_value = np.nan

    if isinstance(x, ArrayPointer):
        def impl(x):
            if len(x) == 0:
                printf("Mean of empty array")
                return zero_value
            return sum(x) / len(x)
        return impl


@expose_and_overload(np.cumsum)
def omnisci_np_cumsum(a):
    if isinstance(a, ArrayPointer):
        eltype = a.eltype

        def impl(a):
            sz = len(a)
            out = Array(sz, eltype)
            out[0] = a[0]
            for i in range(sz):
                out[i] = out[i-1] + a[i]
            return out
        return impl



import math
import numpy as np
from llvmlite import ir
from rbc.utils import get_version
if get_version('numba') >= (0, 49):
    from numba.cpython import mathimpl
    from numba.np import ufunc_db, npyfuncs
    from numba.core import types
else:
    from numba.targets import mathimpl, ufunc_db, npyfuncs
    from numba import types

# tell numba to wire np.exp2 to libm exp2.
mathimpl.unary_math_extern(np.exp2, "exp2f", "exp2")
mathimpl.unary_math_extern(np.log2, "log2f", "log2")
mathimpl.unary_math_extern(math.gamma, "tgammaf", "tgamma")


def np_logaddexp_impl(context, builder, sig, args):
    # based on NumPy impl.
    # https://github.com/numpy/numpy/blob/30f985499b77381ab4748692cc76c55048ca0548/numpy/core/src/npymath/npy_math_internal.h.src#L585-L604
    def impl(x, y):
        if x == y:
            LOGE2 = 0.693147180559945309417232121458176568  # log_e 2
            return x + LOGE2
        else:
            tmp = x - y
            if tmp > 0:
                return x + np.log1p(np.exp(-tmp))
            elif tmp <= 0:
                return y + np.log1p(np.exp(tmp))
            else:
                # NaN's
                return tmp

    return context.compile_internal(builder, impl, sig, args)


def np_logaddexp2_impl(context, builder, sig, args):

    def impl(x, y):
        if x == y:
            return x + 1
        else:
            LOG2E = 1.442695040888963407359924681001892137  # log_2 e
            tmp = x - y
            if tmp > 0:
                return x + (LOG2E * np.log1p(np.exp2(-tmp)))
            elif tmp <= 0:
                return y + (LOG2E * np.log1p(np.exp2(tmp)))
            else:
                # NaN's
                return tmp

    return context.compile_internal(builder, impl, sig, args)


def np_signbit_impl(context, builder, sig, args):
    int64_t = ir.IntType(64)
    double = ir.DoubleType()
    [val] = args
    val = builder.bitcast(builder.fpext(val, double), int64_t)
    res = builder.ashr(val, int64_t(63))
    return res


def np_ldexp_impl(context, builder, sig, args):
    # check ldexp arity
    assert len(args) == 2
    assert len(sig.args) == 2

    dispatch_table = {
        types.float32: 'ldexpf',
        types.float64: 'ldexp',
    }

    return npyfuncs._dispatch_func_by_name_type(context, builder, sig, args,
                                                dispatch_table, 'ldexp')


def np_real_nextafter_impl(context, builder, sig, args):
    npyfuncs._check_arity_and_homogeneity(sig, args, 2)

    dispatch_table = {
        types.float32: 'nextafterf',
        types.float64: 'nextafter',
    }

    return npyfuncs._dispatch_func_by_name_type(context, builder, sig, args,
                                                dispatch_table, 'nextafter')


ufunc_db._lazy_init_db()

# logaddexp
ufunc_db._ufunc_db[np.logaddexp] = {
    'ff->f': np_logaddexp_impl,
    'dd->d': np_logaddexp_impl,
}

# logaddexp2
ufunc_db._ufunc_db[np.logaddexp2] = {
    'ff->f': np_logaddexp2_impl,
    'dd->d': np_logaddexp2_impl,
}

# signbit
ufunc_db._ufunc_db[np.signbit] = {
    'f->?': np_signbit_impl,
    'd->?': np_signbit_impl,
}

# ldexp
ufunc_db._ufunc_db[np.ldexp] = {
    'fi->f': np_ldexp_impl,
    'fl->f': np_ldexp_impl,
    'di->d': np_ldexp_impl,
    'dl->d': np_ldexp_impl,
}

# nextafter
ufunc_db._ufunc_db[np.nextafter] = {
    'ff->f': np_real_nextafter_impl,
    'dd->d': np_real_nextafter_impl,
}



import pytest
import numpy as np
import rbc.omnisci_backend as omni
from rbc.utils import get_version
if get_version('numba') >= (0, 49):
    from numba.core import types
else:
    from numba import types


rbc_omnisci = pytest.importorskip('rbc.omniscidb')
available_version, reason = rbc_omnisci.is_available()
pytestmark = pytest.mark.skipif(not available_version, reason=reason)


@pytest.fixture(scope='module')
def omnisci():
    config = rbc_omnisci.get_client_config(debug=not True)
    m = rbc_omnisci.RemoteOmnisci(**config)
    table_name = 'rbc_test_omnisci_array'
    m.sql_execute('DROP TABLE IF EXISTS {table_name}'.format(**locals()))
    sqltypes = ['FLOAT[]', 'DOUBLE[]',
                'TINYINT[]', 'SMALLINT[]', 'INT[]', 'BIGINT[]',
                'BOOLEAN[]']
    # todo: TEXT ENCODING DICT, TEXT ENCODING NONE, TIMESTAMP, TIME,
    # DATE, DECIMAL/NUMERIC, GEOMETRY: POINT, LINESTRING, POLYGON,
    # MULTIPOLYGON, See
    # https://www.omnisci.com/docs/latest/5_datatypes.html
    colnames = ['f4', 'f8', 'i1', 'i2', 'i4', 'i8', 'b']
    table_defn = ',\n'.join('%s %s' % (n, t)
                            for t, n in zip(sqltypes, colnames))
    m.sql_execute(
        'CREATE TABLE IF NOT EXISTS {table_name} ({table_defn});'
        .format(**locals()))

    def row_value(row, col, colname):
        if colname == 'b':
            return 'ARRAY[%s]' % (', '.join(
                ("'true'" if i % 2 == 0 else "'false'")
                for i in range(-3, 3)))
        if colname.startswith('f'):
            return 'ARRAY[%s]' % (', '.join(
                str(row * 10 + i + 0.5) for i in range(-3, 3)))
        return 'ARRAY[%s]' % (', '.join(
            str(row * 10 + i) for i in range(-3, 3)))

    rows = 5
    for i in range(rows):
        table_row = ', '.join(str(row_value(i, j, n))
                              for j, n in enumerate(colnames))
        m.sql_execute(
            'INSERT INTO {table_name} VALUES ({table_row})'.format(**locals()))
    m.table_name = table_name
    yield m
    try:
        m.sql_execute('DROP TABLE IF EXISTS {table_name}'.format(**locals()))
    except Exception as msg:
        print('%s in deardown' % (type(msg)))


def np_ones(sz):
    return omni.ones(sz, types.int32)


def np_ones_dtype(sz):
    return omni.ones(sz)


def np_ones_like_dtype(i4):
    return omni.ones_like(i4, dtype=types.double)


def np_ones_like(i4):
    return omni.ones_like(i4)


def np_zeros(sz):
    return omni.zeros(sz, np.int32)


def np_zeros_dtype(sz):
    return omni.zeros(sz)


def np_zeros_like(i4):
    return omni.zeros_like(i4)


def np_zeros_like_dtype(i4):
    return omni.zeros_like(i4, dtype=types.double)


def np_full(sz, fill_value):
    return omni.full(sz, fill_value, types.double)


def np_full_dtype(sz, fill_value):
    return omni.full(sz, fill_value)


def np_full_like(i1, fill_value):
    return omni.full_like(i1, fill_value)


def np_full_like_dtype(i1, fill_value):
    return omni.full_like(i1, fill_value, dtype=types.double)


def np_cumsum(sz):
    a = omni.ones(sz)
    return omni.cumsum(a)


array_methods = [
    ('full', 'double[](int64, double)', (5, 3), np.full(5, 3, dtype=np.int32)),
    ('full_dtype', 'double[](int64, double)', (5, 3), np.full(5, 3)),
    ('full_like', 'int8[](int8[], int32)', ('i1', 3), np.full(6, 3, dtype='b')),  # noqa: E501
    ('full_like_dtype', 'double[](int8[], double)', ('i1', 3.0), np.full(6, 3, dtype='q')),  # noqa: E501
    ('ones', 'int32[](int64)', (5,), np.ones(5, dtype=np.int32)),
    ('ones_dtype', 'double[](int64)', (5,), np.ones(5)),
    ('ones_like', 'int32[](int32[])', ('i4',), np.ones(6, dtype='i')),
    ('ones_like_dtype', 'double[](int32[])', ('i4',), np.ones(6, dtype='q')),
    ('zeros', 'int32[](int64)', (5,), np.zeros(5, dtype=np.int32)),
    ('zeros_like', 'int32[](int32[])', ('i4',), np.zeros(6, dtype='i')),
    ('zeros_like_dtype', 'double[](int32[])', ('i4',), np.zeros(6, dtype='q')),
    ('zeros_dtype', 'double[](int64)', (5,), np.zeros(5)),
    ('cumsum', 'double[](int32)', (5,), np.arange(1, 6)),
]


@pytest.mark.parametrize("method, signature, args, expected", array_methods,
                         ids=[item[0] for item in array_methods])
def test_array_methods(omnisci, method, signature, args, expected):
    if (available_version[:3] == (5, 3, 1)
        and method in ['full', 'full_dtype', 'ones', 'ones_dtype', 'zeros',
                       'zeros_dtype', 'cumsum']):
        pytest.skip(
            f'{method}: crashes CPU-only omniscidb server v 5.4 [issue 113]')
    omnisci.reset()

    fn = omnisci(signature)(eval('np_{}'.format(method)))

    query = 'select np_{method}'.format(**locals()) + \
            '(' + ', '.join(map(str, args)) + ')' + \
            ' from {omnisci.table_name};'.format(**locals())

    _, result = omnisci.sql_execute(query)
    out = list(result)[0]

    assert np.array_equal(expected, out[0]), 'np_' + method



import os
from rbc import errors
import numpy as np
import pytest


rbc_omnisci = pytest.importorskip('rbc.omniscidb')
available_version, reason = rbc_omnisci.is_available()
# Throw an error on Travis CI if the server is not available
if "TRAVIS" in os.environ and not available_version:
    pytest.exit(msg=reason, returncode=1)
pytestmark = pytest.mark.skipif(not available_version, reason=reason)


def test_get_client_config(tmpdir):
    d = tmpdir.mkdir("omnisci")
    fh = d.join("client.conf")
    fh.write("""
    [user]
name  =  foo
password = secret

[server]
port: 1234
host: example.com

[rbc]
debug: False
use_host_target: False
# server: Server [NOT IMPL]
# target_info: TargetInfo
""")
    conf_file = os.path.join(fh.dirname, fh.basename)

    old_conf = os.environ.get('OMNISCI_CLIENT_CONF')
    os.environ['OMNISCI_CLIENT_CONF'] = conf_file

    try:
        conf = rbc_omnisci.get_client_config()
        assert conf['user'] == 'foo'
        assert conf['password'] == 'secret'
        assert conf['port'] == 1234
        assert conf['host'] == 'example.com'
        assert conf['dbname'] == 'omnisci'
        assert conf['debug'] == bool(0)
        conf = rbc_omnisci.get_client_config(dbname='test')
        assert conf['dbname'] == 'test'
    finally:
        if old_conf is None:
            del os.environ['OMNISCI_CLIENT_CONF']
        else:
            os.environ['OMNISCI_CLIENT_CONF'] = old_conf


@pytest.fixture(scope='module')
def nb_version():
    from rbc.utils import get_version
    return get_version('numba')


@pytest.fixture(scope='module')
def omnisci():
    config = rbc_omnisci.get_client_config(debug=not True)
    m = rbc_omnisci.RemoteOmnisci(**config)
    table_name = 'rbc_test_omnisci'
    m.sql_execute('DROP TABLE IF EXISTS {table_name}'.format(**locals()))
    sqltypes = ['FLOAT', 'DOUBLE', 'TINYINT', 'SMALLINT', 'INT', 'BIGINT',
                'BOOLEAN']
    # todo: TEXT ENCODING DICT, TEXT ENCODING NONE, TIMESTAMP, TIME,
    # DATE, DECIMAL/NUMERIC, GEOMETRY: POINT, LINESTRING, POLYGON,
    # MULTIPOLYGON, See
    # https://www.omnisci.com/docs/latest/5_datatypes.html
    colnames = ['f4', 'f8', 'i1', 'i2', 'i4', 'i8', 'b']
    table_defn = ',\n'.join('%s %s' % (n, t)
                            for t, n in zip(sqltypes, colnames))
    m.sql_execute(
        'CREATE TABLE IF NOT EXISTS {table_name} ({table_defn});'
        .format(**locals()))

    def row_value(row, col, colname):
        if colname == 'b':
            return ("'true'" if row % 2 == 0 else "'false'")
        return row

    rows = 5
    for i in range(rows):
        table_row = ', '.join(str(row_value(i, j, n))
                              for j, n in enumerate(colnames))
        m.sql_execute(
            'INSERT INTO {table_name} VALUES ({table_row})'.format(**locals()))
    m.table_name = table_name
    yield m
    m.sql_execute('DROP TABLE IF EXISTS {table_name}'.format(**locals()))


def test_redefine(omnisci):
    omnisci.reset()

    @omnisci('i32(i32)')
    def incr(x):
        return x + 1

    desrc, result = omnisci.sql_execute(
        'select i4, incr(i4) from {omnisci.table_name}'.format(**locals()))
    for x, x1 in result:
        assert x1 == x + 1

    # Re-defining triggers a warning message when in debug mode
    @omnisci('i32(i32)')  # noqa: F811
    def incr(x):  # noqa: F811
        return x + 2

    desrc, result = omnisci.sql_execute(
        'select i4, incr(i4) from {omnisci.table_name}'.format(**locals()))
    for x, x1 in result:
        assert x1 == x + 2


def test_forbidden_define(omnisci):
    if omnisci.version > (5, 1):
        pytest.skip(
            f'forbidden defines not required for OmnisciDB {omnisci.version}')

    omnisci.reset()

    msg = "Attempt to define function with name `{name}`"

    @omnisci('double(double)')
    def sinh(x):
        return np.sinh(x)

    with pytest.raises(errors.ForbiddenNameError) as excinfo:
        omnisci.register()
    assert msg.format(name='sinh') in str(excinfo.value)

    omnisci.reset()

    @omnisci('double(double)')
    def trunc(x):
        return np.trunc(x)

    with pytest.raises(errors.ForbiddenNameError) as excinfo:
        omnisci.register()
    assert msg.format(name='trunc') in str(excinfo.value)

    omnisci.reset()


def test_single_argument_overloading(omnisci):
    omnisci.reset()

    @omnisci(
        'f64(f64)',
        'i64(i64)',
        'i32(i32)',
        'f32(f32)',
        # 'i32(f32)',
    )
    def mydecr(x):
        return x - 1
    desrc, result = omnisci.sql_execute(
        'select f4, mydecr(f4) from {omnisci.table_name}'.format(**locals()))
    result = list(result)
    assert len(result) > 0
    for x, x1 in result:
        assert x1 == x - 1
        assert isinstance(x1, type(x))
    desrc, result = omnisci.sql_execute(
        'select f8, mydecr(f8) from {omnisci.table_name}'.format(**locals()))
    result = list(result)
    assert len(result) > 0
    for x, x1 in result:
        assert x1 == x - 1
        assert isinstance(x1, type(x))
    desrc, result = omnisci.sql_execute(
        'select i4, mydecr(i4) from {omnisci.table_name}'.format(**locals()))
    result = list(result)
    assert len(result) > 0
    for x, x1 in result:
        assert x1 == x - 1
        assert isinstance(x1, type(x))


def test_numpy_forbidden_ufunc(omnisci, nb_version):
    omnisci.reset()

    if omnisci.version >= (5, 5) and nb_version >= (0, 52):
        pytest.skip(
            f'forbidden ufunc not required for OmniSciDB {omnisci.version} '
            f'and Numba {nb_version}')

    if not omnisci.has_cuda:
        pytest.skip('forbidden ufunc not required for CUDA-disabled OmniSciDB')

    if omnisci.version >= (5, 5):
        # Requires: https://github.com/omnisci/omniscidb-internal/pull/4955
        pytest.skip('forbidden ufunc not required if CPU ufuncs work [omniscidb-interal PR 4955]')

    msg = "Attempt to use function with name `{ufunc}`"

    @omnisci('double(double)')
    def arcsin(x):
        return np.arcsin(x)

    with pytest.raises(errors.ForbiddenIntrinsicError) as excinfo:
        omnisci.register()
    assert msg.format(ufunc='asin') in str(excinfo.value)

    omnisci.reset()

    @omnisci('float32(float32, float32)')
    def logaddexp(x, y):
        return np.logaddexp(x, y)

    with pytest.raises(errors.ForbiddenIntrinsicError) as excinfo:
        omnisci.register()
    assert msg.format(ufunc='log1pf') in str(excinfo.value)

    omnisci.reset()


def test_thrift_api_doc(omnisci):
    omnisci.reset()

    @omnisci('double(int, double)',
             'float(int, float)',
             'int(int, int)')
    def foo(i, v):
        return v * i + 55

    descr, result = omnisci.sql_execute(
        'select f8, foo(i4, f8) from {omnisci.table_name}'.format(**locals()))
    result = list(result)
    assert len(result) > 0
    for i, (x, x1) in enumerate(result):
        assert x1 == x * i + 55
        assert isinstance(x1, type(x))


def test_manual_ir(omnisci):
    omnisci.reset()
    descr, result = omnisci.sql_execute(
        'SELECT * FROM {omnisci.table_name}'.format(**locals()))
    result = list(result)
    assert result == [(0.0, 0.0, 0, 0, 0, 0, 1), (1.0, 1.0, 1, 1, 1, 1, 0),
                      (2.0, 2.0, 2, 2, 2, 2, 1), (3.0, 3.0, 3, 3, 3, 3, 0),
                      (4.0, 4.0, 4, 4, 4, 4, 1)]
    device_params = omnisci.thrift_call('get_device_parameters',
                                        omnisci.session_id)
    cpu_target_triple = device_params['cpu_triple']
    cpu_target_datalayout = "e-m:e-i64:64-f80:128-n8:16:32:64-S128"
    gpu_target_triple = device_params.get('gpu_triple')
    gpu_target_datalayout = ("e-p:64:64:64-i1:8:8-i8:8:8-"
                             "i16:16:16-i32:32:32-i64:64:64-"
                             "f32:32:32-f64:64:64-v16:16:16-"
                             "v32:32:32-v64:64:64-v128:128:128-n16:32:64")

    foo_ir = '''\
define i32 @foobar(i32 %.1, i32 %.2) {
entry:
  %.18.i = mul i32 %.2, %.1
  %.33.i = add i32 %.18.i, 55
  ret i32 %.33.i
}
'''
    ast_signatures = "foobar 'int32(int32, int32)'"
    device_ir_map = dict()
    device_ir_map['cpu'] = '''
target datalayout = "{cpu_target_datalayout}"
target triple = "{cpu_target_triple}"
{foo_ir}
'''.format(**locals())

    if gpu_target_triple is not None:
        device_ir_map['gpu'] = '''
target datalayout = "{gpu_target_datalayout}"
target triple = "{gpu_target_triple}"
{foo_ir}
'''.format(**locals())

    omnisci.thrift_call('register_runtime_udf', omnisci.session_id,
                        ast_signatures, device_ir_map)
    omnisci._last_ir_map = {}  # hack
    descr, result = omnisci.sql_execute(
        'SELECT i4, foobar(i4, i4) FROM {omnisci.table_name}'
        .format(**locals()))
    result = list(result)
    assert len(result) > 0
    for x, r in result:
        assert r == x * x + 55


def test_ir_parse_error(omnisci):
    device_params = omnisci.thrift_call('get_device_parameters',
                                        omnisci.session_id)
    foo_ir = '''\
define i32 @foobar(i32 %.1, i32 %.2) {
entry:
  %.18.i = mul i32 %.2, %.1
  %.33.i = add i32 %.18.i, 55
  ret i32 %.33.i

'''
    ast_signatures = "foobar 'int32(int32, int32)'"
    device_ir_map = dict()
    device_ir_map['cpu'] = foo_ir

    gpu_target_triple = device_params.get('gpu_triple')
    if gpu_target_triple is not None:
        device_ir_map['gpu_triple'] = foo_ir

    with pytest.raises(Exception, match=r".*LLVM IR ParseError:"):
        omnisci.thrift_call('register_runtime_udf', omnisci.session_id,
                            ast_signatures, device_ir_map)


def test_ir_query_error(omnisci):
    pytest.skip("requires omniscidb-internal catching undefined symbols")

    device_params = omnisci.thrift_call('get_device_parameters',
                                        omnisci.session_id)
    gpu_target_triple = device_params.get('gpu_triple')
    foo_ir = '''\
define i32 @foobarrr(i32 %.1, i32 %.2) {
entry:
  %.18.i = mul i32 %.2, %.1
  %.33.i = add i32 %.18.i, 55
  ret i32 %.33.i
}
'''
    ast_signatures = "foobar 'int32(int32, int32)'"
    device_ir_map = dict()
    device_ir_map['cpu'] = foo_ir
    if gpu_target_triple is not None:
        device_ir_map['gpu'] = foo_ir

    omnisci.thrift_call('register_runtime_udf', omnisci.session_id,
                        ast_signatures, device_ir_map)
    try:
        omnisci.sql_execute(
            'SELECT i4, foobar(i4, i4) FROM {omnisci.table_name}'
            .format(**locals()))
    except errors.OmnisciServerError as msg:
        assert "use of undefined value '@foobar'" in str(msg)


def test_multiple_implementation(omnisci):
    omnisci.reset()

    @omnisci('int(f64)', 'int(i64)')  # noqa: F811
    def bits(x):
        return 64

    @omnisci('int(f32)', 'int(i32)')  # noqa: F811
    def bits(x):  # noqa: F811
        return 32

    @omnisci('int(i16)')  # noqa: F811
    def bits(x):  # noqa: F811
        return 16

    @omnisci('int(i8)')  # noqa: F811
    def bits(x):  # noqa: F811
        return 8

    descr, result = omnisci.sql_execute(
        'select bits(i1), bits(i2), bits(i4), bits(f4), bits(i8), bits(f8)'
        ' from {omnisci.table_name} limit 1'.format(**locals()))
    result = list(result)
    assert len(result) == 1
    assert result[0] == (8, 16, 32, 32, 64, 64)


def test_loadtime_udf(omnisci):
    # This test requires that omnisci server is started with `--udf
    # sample_udf.cpp` option where the .cpp file defines `udf_diff`
    try:
        desrc, result = omnisci.sql_execute(
            'select i4, udf_diff2(i4, i4) from {omnisci.table_name}'
            .format(**locals()))
    except Exception as msg:
        assert 'No match found for function signature udf_diff' in str(msg)
        return
    result = list(result)
    for i_, (i, d) in enumerate(result):
        assert i_ == i
        assert d == 0


def test_f32(omnisci):
    """If UDF name ends with an underscore, expect strange behaviour. For
    instance, defining

      @omnisci('f32(f32)', 'f32(f64)')
      def f32_(x): return x+4.5

    the query `select f32_(0.0E0))` fails but not when defining

      @omnisci('f32(f64)', 'f32(f32)')
      def f32_(x): return x+4.5

    (notice the order of signatures in omnisci decorator argument).
    """

    @omnisci('f32(f32)', 'f32(f64)')  # noqa: F811
    def f_32(x): return x+4.5
    descr, result = omnisci.sql_execute(
        'select f_32(0.0E0) from {omnisci.table_name} limit 1'
        .format(**locals()))
    assert list(result)[0] == (4.5,)


def test_castop(omnisci):
    @omnisci('i16(i16)')  # noqa: F811
    def i32(x): return x+2  # noqa: F811

    @omnisci('i32(i32)')  # noqa: F811
    def i32(x): return x+4  # noqa: F811

    descr, result = omnisci.sql_execute(
        'select i32(cast(1 as int))'
        ' from {omnisci.table_name} limit 1'.format(**locals()))
    assert list(result)[0] == (5,)


def test_casting(omnisci):
    """
    Define UDFs:
      i8(<tinyint>), i16(<smallint>), i32(<int>), i64(<bigint>)
      f32(<float>), f64(<double>)

    The following table defines the behavior of applying these UDFs to
    values with different types:

             | Functions applied to <itype value>
    itype    | i8   | i16  | i32  | i64  | f32  | f64  |
    ---------+------+------+------+------+------+------+
    tinyint  | OK   | OK   | OK   | OK   | FAIL | FAIL |
    smallint | FAIL | OK   | OK   | OK   | FAIL | FAIL |
    int      | FAIL | FAIL | OK   | OK   | FAIL | FAIL |
    bigint   | FAIL | FAIL | FAIL | OK   | FAIL | FAIL |
    float    | FAIL | FAIL | FAIL | FAIL | OK   | OK   |
    double   | FAIL | FAIL | FAIL | FAIL | FAIL | OK   |
    """
    omnisci.reset()

    @omnisci('i8(i8)')    # noqa: F811
    def i8(x): return x+1

    @omnisci('i16(i16)')  # noqa: F811
    def i16(x): return x+2

    @omnisci('i32(i32)')  # noqa: F811
    def i32(x): return x+4

    @omnisci('i64(i64)')  # noqa: F811
    def i64(x): return x+8

    @omnisci('f32(f32)')  # noqa: F811
    def f32(x): return x+4.5

    # cannot create a 8-bit and 16-bit int literals in sql, so, using
    # the following helper functions:
    @omnisci('i8(i8)', 'i8(i16)', 'i8(i32)')    # noqa: F811
    def i_8(x): return x+1

    @omnisci('i16(i16)', 'i16(i32)')  # noqa: F811
    def i_16(x): return x+2

    # cannot create a 32-bit float literal in sql, so, using a helper
    # function for that:
    @omnisci('f32(f32)', 'f32(f64)')  # noqa: F811
    def f_32(x): return x+4.5

    @omnisci('f64(f64)')  # noqa: F811
    def f64(x): return x+8.5

    @omnisci('i8(i8)')
    def ifoo(x): return x + 1

    @omnisci('i16(i16)')  # noqa: F811
    def ifoo(x): return x + 2  # noqa: F811

    @omnisci('i32(i32)')  # noqa: F811
    def ifoo(x): return x + 4  # noqa: F811

    @omnisci('i64(i64)')  # noqa: F811
    def ifoo(x): return x + 8  # noqa: F811

    @omnisci('f32(f32)')
    def ffoo(x): return x + 4.5  # noqa: F811

    @omnisci('f64(f64)')  # noqa: F811
    def ffoo(x): return x + 8.5  # noqa: F811

    descr, result = omnisci.sql_execute(
        'select i_8(0),i_16(0),i32(0),i64(0) from {omnisci.table_name} limit 1'
        .format(**locals()))
    assert list(result)[0] == (1, 2, 4, 8)

    descr, result = omnisci.sql_execute(
        'select i_8(i1),i16(i2),i32(i4),i64(i8)'
        ' from {omnisci.table_name} limit 1'
        .format(**locals()))
    assert list(result)[0] == (1, 2, 4, 8)

    descr, result = omnisci.sql_execute(
        'select ifoo(i_8(0)),ifoo(i_16(0)),ifoo(i32(0)),ifoo(i64(0))'
        ' from {omnisci.table_name} limit 1'
        .format(**locals()))
    assert list(result)[0] == (1+1, 2+2, 4+4, 8+8)

    descr, result = omnisci.sql_execute(
        'select ifoo(i1),ifoo(i2),ifoo(i4),ifoo(i8)'
        ' from {omnisci.table_name} limit 1'
        .format(**locals()))
    assert list(result)[0] == (1, 2, 4, 8)

    descr, result = omnisci.sql_execute(
        'select i64(i_8(0)), i64(i_16(0)),i64(i32(0)),i64(i64(0))'
        ' from {omnisci.table_name} limit 1'.format(**locals()))
    assert list(result)[0] == (9, 10, 12, 16)

    descr, result = omnisci.sql_execute(
        'select i64(i1), i64(i2),i64(i4),i64(i8)'
        ' from {omnisci.table_name} limit 1'.format(**locals()))
    assert list(result)[0] == (8, 8, 8, 8)

    descr, result = omnisci.sql_execute(
        'select i32(i_8(0)), i32(i_16(0)),i32(i32(0))'
        ' from {omnisci.table_name} limit 1'.format(**locals()))
    assert list(result)[0] == (5, 6, 8)

    descr, result = omnisci.sql_execute(
        'select i32(i1), i32(i2),i32(i4)'
        ' from {omnisci.table_name} limit 1'.format(**locals()))
    assert list(result)[0] == (4, 4, 4)

    descr, result = omnisci.sql_execute(
        'select i16(i_8(0)), i16(i_16(0)) from {omnisci.table_name} limit 1'
        .format(**locals()))
    assert list(result)[0] == (3, 4)

    descr, result = omnisci.sql_execute(
        'select i16(i1), i16(i2) from {omnisci.table_name} limit 1'
        .format(**locals()))
    assert list(result)[0] == (2, 2)

    descr, result = omnisci.sql_execute(
        'select i8(i_8(0)) from {omnisci.table_name} limit 1'
        .format(**locals()))
    assert list(result)[0] == (2,)

    descr, result = omnisci.sql_execute(
        'select i8(i1) from {omnisci.table_name} limit 1'.format(**locals()))
    assert list(result)[0] == (1,)

    descr, result = omnisci.sql_execute(
        'select f_32(0.0E0), f64(0.0E0) from {omnisci.table_name} limit 1'
        .format(**locals()))
    assert list(result)[0] == (4.5, 8.5)

    descr, result = omnisci.sql_execute(
        'select f_32(f4), f64(f8) from {omnisci.table_name} limit 1'
        .format(**locals()))
    assert list(result)[0] == (4.5, 8.5)

    descr, result = omnisci.sql_execute(
        'select f64(f_32(0.0E0)), f64(f64(0.0E0))'
        ' from {omnisci.table_name} limit 1'.format(**locals()))
    assert list(result)[0] == (13.0, 17.0)

    descr, result = omnisci.sql_execute(
        'select f64(f4), f64(f8)'
        ' from {omnisci.table_name} limit 1'.format(**locals()))
    assert list(result)[0] == (8.5, 8.5)

    descr, result = omnisci.sql_execute(
        'select f32(f_32(0.0E0)) from {omnisci.table_name} limit 1'
        .format(**locals()))
    assert list(result)[0] == (9.0,)

    descr, result = omnisci.sql_execute(
        'select f32(f4) from {omnisci.table_name} limit 1'
        .format(**locals()))
    assert list(result)[0] == (4.5,)

    descr, result = omnisci.sql_execute(
        'select ffoo(f_32(0.0E0)), ffoo(f64(0.0E0))'
        ' from {omnisci.table_name} limit 1'.format(**locals()))
    assert list(result)[0] == (9.0, 17.0)

    descr, result = omnisci.sql_execute(
        'select ffoo(f4), ffoo(f8)'
        ' from {omnisci.table_name} limit 1'.format(**locals()))
    assert list(result)[0] == (4.5, 8.5)

    for v, f, t in [
            ('i64(0)', r'i32', r'BIGINT'),
            ('i64(0)', r'i16', r'BIGINT'),
            ('i64(0)', r'i8', r'BIGINT'),
            ('i8', r'i32', r'BIGINT'),
            ('i8', r'i16', r'BIGINT'),
            ('i8', r'i8', r'BIGINT'),
            ('i32(0)', r'i16', r'INTEGER'),
            ('i32(0)', r'i8', r'INTEGER'),
            ('i4', r'i16', r'INTEGER'),
            ('i4', r'i8', r'INTEGER'),
            ('i_16(0)', r'i8', r'SMALLINT'),
            ('i2', r'i8', r'SMALLINT'),
            ('f64(0)', r'f32', r'DOUBLE'),
            ('f8', r'f32', r'DOUBLE'),
    ]:
        q = ('select '+f+'('+v+') from {omnisci.table_name} limit 1'
             .format(**locals()))
        match = (r".*(Function "+f+r"\("+t+r"\) not supported"
                 r"|Could not bind "+f+r"\("+t+r"\))")
        with pytest.raises(Exception, match=match):
            descr, result = omnisci.sql_execute(q)
            print('query: ', q)
            print('expected: ', match)

    for f in [r'f64', r'f32']:
        for at, av, r in [
                (r'BIGINT', 'i64(0)', (16.5,)),
                (r'INTEGER', 'i32(0)', (12.5,)),
                (r'SMALLINT', 'i_16(0)', (10.5,)),
                (r'TINYINT', 'i_8(0)', (9.5,)),
                (r'BIGINT', 'i8', (8.5,)),
                (r'INTEGER', 'i4', (8.5,)),
                (r'SMALLINT', 'i2', (8.5,)),
                (r'TINYINT', 'i1', (8.5,)),
        ]:
            if f == r'f64':  # temporary: allow integers as double arguments
                descr, result = omnisci.sql_execute(
                    'select '+f+'('+av+') from {omnisci.table_name} limit 1'
                    .format(**locals()))
                assert list(result)[0] == r
                continue
            with pytest.raises(
                    Exception,
                    match=(r".*(Function "+f+r"\("+at+r"\) not supported"
                           r"|Could not bind "+f+r"\("+at+r"\))")):
                descr, result = omnisci.sql_execute(
                    'select '+f+'('+av+') from {omnisci.table_name} limit 1'
                    .format(**locals()))

    for f in [r'i64', r'i32', r'i16', r'i8']:
        for at, av in [
                (r'DOUBLE', 'f64(0E0)'),
                (r'FLOAT', 'f_32(0E0)'),
                (r'DOUBLE', 'f8'),
                (r'FLOAT', 'f4'),
        ]:
            with pytest.raises(
                    Exception,
                    match=(r".*(Function "+f+r"\("+at+r"\) not supported"
                           r"|Could not bind "+f+r"\("+at+r"\))")):
                descr, result = omnisci.sql_execute(
                    'select '+f+'('+av+') from {omnisci.table_name} limit 1'
                    .format(**locals()))


def test_truncate_issue(omnisci):
    omnisci.reset()

    @omnisci('int(f64)', 'int(i64)')  # noqa: F811
    def bits(x):  # noqa: F811
        return 64

    @omnisci('int(f32)', 'int(i32)')  # noqa: F811
    def bits(x):  # noqa: F811
        return 32

    @omnisci('int(i16)')  # noqa: F811
    def bits(x):  # noqa: F811
        return 16

    @omnisci('int(i8)')  # noqa: F811
    def bits(x):  # noqa: F811
        return 8

    descr, result = omnisci.sql_execute(
        'select bits(truncate(2016, 1)) from {omnisci.table_name} limit 1'
        .format(**locals()))
    assert list(result)[0] in [(16,), (32,)]

    descr, result = omnisci.sql_execute(
        'select bits(truncate(cast(2016.0 as smallint), 1))'
        ' from {omnisci.table_name} limit 1'
        .format(**locals()))
    assert list(result)[0] == (16,)

    descr, result = omnisci.sql_execute(
        'select bits(truncate(cast(2016.0 as int), 1))'
        ' from {omnisci.table_name} limit 1'
        .format(**locals()))
    assert list(result)[0] == (32,)

    descr, result = omnisci.sql_execute(
        'select bits(truncate(cast(2016.0 as bigint), 1))'
        ' from {omnisci.table_name} limit 1'
        .format(**locals()))
    assert list(result)[0] == (64,)

    descr, result = omnisci.sql_execute(
        'select bits(truncate(cast(2016.0 as float), 1))'
        ' from {omnisci.table_name} limit 1'
        .format(**locals()))
    assert list(result)[0] == (32,)

    descr, result = omnisci.sql_execute(
        'select bits(truncate(cast(2016.0 as double), 1))'
        ' from {omnisci.table_name} limit 1'
        .format(**locals()))
    assert list(result)[0] == (64,)



from numba.core.imputils import Registry
from numba.core import typing
from numba.cuda import libdevice, mathimpl
from numba.types import float32, float64, int64, uint64, int32, boolean
import numpy as np

registry = Registry()
lower = registry.lower

# The three lists below - booleans, unarys, binarys - maps numpy ufuncs
# to libdevice intrinsics. Some functions are commented out because either
# the CPU version works or because there is no 1:1 corresponding from the
# numpy ufunc to libdevice. For instance, there are 12 intrinsic on
# libdevice to add two numbers.

booleans = []
booleans += [('isnand', 'isnanf', np.isnan)]
booleans += [('isinfd', 'isinff', np.isinf)]
booleans += [('isfinited', 'finitef', np.isfinite)]
booleans += [('isnand', 'isnanf', np.isnat)]

unarys = []
unarys += [('fabs', 'fabsf', np.fabs)]
unarys += [('exp', 'expf', np.exp)]
unarys += [('exp2', 'exp2f', np.exp2)]
unarys += [('log', 'logf', np.log)]
unarys += [('log2', 'log2f', np.log2)]
unarys += [('log10', 'log10f', np.log10)]
unarys += [('expm1', 'expm1f', np.expm1)]
unarys += [('log1p', 'log1pf', np.log1p)]
unarys += [('sqrt', 'sqrtf', np.sqrt)]
unarys += [('cbrt', 'cbrtf', np.cbrt)]
unarys += [('sin', 'sinf', np.sin)]
unarys += [('cos', 'cosf', np.cos)]
unarys += [('tan', 'tanf', np.tan)]
unarys += [('asin', 'asinf', np.arcsin)]
unarys += [('acos', 'acosf', np.arccos)]
unarys += [('atan', 'atanf', np.arctan)]
unarys += [('sinh', 'sinhf', np.sinh)]
unarys += [('cosh', 'coshf', np.cosh)]
unarys += [('tanh', 'tanhf', np.tanh)]
unarys += [('asinh', 'asinhf', np.arcsinh)]
unarys += [('acosh', 'acoshf', np.arccosh)]
unarys += [('atanh', 'atanhf', np.arctanh)]
unarys += [('fabs', 'fabsf', np.fabs)]
unarys += [('floor', 'floorf', np.floor)]
unarys += [('ceil', 'ceilf', np.ceil)]
unarys += [('trunc', 'truncf', np.trunc)]
# unarys += [(np.negative)]
# unarys += [(np.positive)]
# unarys += [(np.absolute)]
# unarys += [(np.rint)]
# unarys += [(np.sign)]
# unarys += [(np.conj)]
# unarys += [(np.conjugate)]
# unarys += [(np.square)]
# unarys += [(np.reciprocal)]
# unarys += [(np.degrees)]
# unarys += [(np.radians)]
# unarys += [(np.deg2rad)]
# unarys += [(np.rad2deg)]
# unarys += [(np.invert)]
# unarys += [('signbitd', 'signbitf', np.signbit)]
# unarys += [(np.spacing)]

binarys = []
binarys += [('pow', 'powf', np.power)]
binarys += [('fmod', 'fmodf', np.fmod)]
binarys += [('atan2', 'atan2f', np.arctan2)]
binarys += [('hypot', 'hypotf', np.hypot)]
binarys += [('fmax', 'fmaxf', np.fmax)]
binarys += [('fmin', 'fminf', np.fmin)]
binarys += [('copysign', 'copysignf', np.copysign)]
binarys += [('nextafter', 'nextafterf', np.nextafter)]
binarys += [('fmod', 'fmodf', np.fmod)]
# binarys += [(np.add)]
# binarys += [(np.subtract)]
# binarys += [(np.multiply)]
# binarys += [(np.matmul)]
# binarys += [(np.divide)]
# binarys += [(np.logaddexp)]
# binarys += [(np.logaddexp2)]
# binarys += [(np.true_divide)]
# binarys += [(np.floor_divide)]
# binarys += [(np.float_power)]
# binarys += [('remainder', 'remainderf', np.remainder)]
# binarys += [('remainder', 'remainderf', np.mod)]
# binarys += [(np.divmod)]
# binarys += [(np.heaviside)]
# binarys += [(np.gcd)]
# binarys += [(np.lcm)]
# binarys += [(np.bitwise_and)]
# binarys += [(np.bitwise_or)]
# binarys += [(np.left_shift)]
# binarys += [(np.right_shift)]
# binarys += [(np.greater)]
# binarys += [(np.greater_equal)]
# binarys += [(np.less)]
# binarys += [(np.less_equal)]
# binarys += [(np.not_equal)]
# binarys += [(np.equal)]
# binarys += [(np.logical_and)]
# binarys += [(np.logical_or)]
# binarys += [(np.logical_xor)]
# binarys += [(np.maximum)]
# binarys += [(np.minimum)]

for fname64, fname32, key in booleans:
    impl32 = getattr(libdevice, fname32)
    impl64 = getattr(libdevice, fname64)
    mathimpl.impl_boolean(key, float32, impl32)
    mathimpl.impl_boolean(key, float64, impl64)

for fname64, fname32, key in unarys:
    impl32 = getattr(libdevice, fname32)
    impl64 = getattr(libdevice, fname64)
    mathimpl.impl_unary(key, float32, impl32)
    mathimpl.impl_unary(key, float64, impl64)
    mathimpl.impl_unary_int(key, int64, impl64)
    mathimpl.impl_unary_int(key, uint64, impl64)

for fname64, fname32, key in binarys:
    impl32 = getattr(libdevice, fname32)
    impl64 = getattr(libdevice, fname64)
    mathimpl.impl_binary(key, float32, impl32)
    mathimpl.impl_binary(key, float64, impl64)
    mathimpl.impl_binary_int(key, int64, impl64)
    mathimpl.impl_binary_int(key, uint64, impl64)


# np.signbit
def impl_signbit(ty, libfunc):
    def lower_signbit_impl(context, builder, sig, args):
        signbit_sig = typing.signature(int32, ty)  # (ty) -> int32
        libfunc_impl = context.get_function(libfunc, signbit_sig)
        result = libfunc_impl(builder, args)
        return context.cast(builder, result, int32, boolean)

    lower(np.signbit, ty)(lower_signbit_impl)


impl_signbit(float32, libdevice.signbitf)
impl_signbit(float64, libdevice.signbitd)


# np.ldexp
def impl_ldexp(ty, libfunc):
    def lower_ldexp_impl(context, builder, sig, args):
        ldexp_sig = typing.signature(ty, ty, int32)  # (ty, int32) -> ty
        libfunc_impl = context.get_function(libfunc, ldexp_sig)
        return libfunc_impl(builder, args)

    lower(np.ldexp, ty, int32)(lower_ldexp_impl)


impl_ldexp(float32, libdevice.ldexpf)
impl_ldexp(float64, libdevice.ldexp)


# np.logaddexp and np.logaddexp2
def impl_logaddexp(ty):
    from .npy_mathimpl import np_logaddexp_impl, np_logaddexp2_impl
    lower(np.logaddexp, ty, ty)(np_logaddexp_impl)
    lower(np.logaddexp2, ty, ty)(np_logaddexp2_impl)


impl_logaddexp(float32)
impl_logaddexp(float64)



import pytest
from rbc.tests import omnisci_fixture


@pytest.fixture(scope='module')
def omnisci():

    for o in omnisci_fixture(globals(), minimal_version=(5, 5)):
        define(o)
        yield o


def define(omnisci):

    @omnisci('int32(Column<int64>, RowMultiplier, OutputColumn<int64>)')
    def text_rbc_copy_c_rowmul(x, m, y):
        for i in range(len(x)):
            # TODO: implement is_null support
            y[i] = x[i]
        return len(x)

    @omnisci('int32(Cursor<int64, double>, RowMultiplier,'
             ' OutputColumn<int64>, OutputColumn<double>)')
    def text_rbc_copy_cc_rowmul(x, x2, m, y, y2):
        for i in range(len(x)):
            y[i] = x[i]
            y2[i] = x2[i]
        return len(x)

    @omnisci('int32(Cursor<int64>, Cursor<double>, RowMultiplier,'
             ' OutputColumn<int64>, OutputColumn<double>)')
    def text_rbc_copy_c_c_rowmul(x, x2, m, y, y2):
        for i in range(len(x)):
            y[i] = x[i]
            y2[i] = x2[i]
        return len(x)

    @omnisci('int32(Cursor<int32>, Cursor<double, int64>, RowMultiplier,'
             ' OutputColumn<int32>, OutputColumn<double>, OutputColumn<int64>)')
    def text_rbc_copy_c_cc_rowmul(x, x2, x3, m, y, y2, y3):
        for i in range(len(x)):
            y[i] = x[i]
            y2[i] = x2[i]
            y3[i] = x3[i]
        return len(x)

    @omnisci('int32(Cursor<int32, int64>, Cursor<double>, RowMultiplier,'
             ' OutputColumn<int32>, OutputColumn<int64>, OutputColumn<double>)')
    def text_rbc_copy_cc_c_rowmul(x, x2, x3, m, y, y2, y3):
        for i in range(len(x)):
            y[i] = x[i]
            y2[i] = x2[i]
            y3[i] = x3[i]
        return len(x)

    @omnisci('int32(Cursor<int32, int64>, Cursor<double, float>, RowMultiplier,'
             ' OutputColumn<int32>, OutputColumn<int64>, OutputColumn<double>,'
             ' OutputColumn<float>)')
    def text_rbc_copy_cc_cc_rowmul(x, x2, x3, x4, m, y, y2, y3, y4):
        for i in range(len(x)):
            y[i] = x[i]
            y2[i] = x2[i]
            y3[i] = x3[i]
            y4[i] = x4[i]
        return len(x)

    @omnisci('int32(Cursor<int32, int64, double, float>, RowMultiplier,'
             ' OutputColumn<int32>, OutputColumn<int64>, OutputColumn<double>,'
             ' OutputColumn<float>)')
    def text_rbc_copy_cccc_rowmul(x, x2, x3, x4, m, y, y2, y3, y4):
        for i in range(len(x)):
            y[i] = x[i]
            y2[i] = x2[i]
            y3[i] = x3[i]
            y4[i] = x4[i]
        return len(x)


@pytest.mark.parametrize("inputs",
                         ['i8', 'cursor(i8,f8)', 'i8;f8', 'i4;cursor(f8,i8)',
                          'cursor(i4,i8);f8', 'cursor(i4,i8);cursor(f8,f4)',
                          'cursor(i4,i8,f8,f4)'])
def test_copy(omnisci, inputs):
    omnisci.require_version((5, 5), 'Requires omniscidb-internal PR 5134')

    groups = inputs.split(';')
    table_names = [f'{omnisci.table_name}'] * len(groups)

    group_results = {}
    args = []
    cc = []
    expected = [()] * 5
    for colnames, table_name in zip(groups, table_names):
        if colnames.startswith('cursor('):
            colnames = colnames[7:-1]
        result = group_results.get(colnames)
        if result is None:
            _, result = omnisci.sql_execute(f'select {colnames} from {table_name}')
            result = group_results[colnames] = list(result)
        expected = [row1 + row2 for row1, row2 in zip(expected, result)]
        args.append(f'cursor(select {colnames} from {table_name})')
        cc.append((colnames.count(',') + 1) * 'c')
    args.append('1')
    args = ', '.join(args)
    cc = '_'.join(cc)

    query = f'select * from table(text_rbc_copy_{cc}_rowmul({args}))'
    _, result = omnisci.sql_execute(query)
    result = list(result)

    assert result == expected
